
"""    
    This package contains implementations of linear algebra operations for sparse and dense matrices and
    utils for convenient operation of nimfa library.
"""

from . import linalg
from . import utils
