
"""    
    This module lists data sets needed to run Nimfa's use cases.

    .. note:: Relevant code for data processing is provided within each example. To run an example download the relevant data and put it in ``nimfa/datasets`` directory.
    
    Data sets:
    
        #. Leukemia data set; `leukemia gene expressions`_, `leukemia samples`_, `leukemia genes`_.
        #. Medulloblastoma data set; `medulloblastoma gene expressions`_, `medulloblastoma samples`_, `medulloblastoma genes`_.
        #. `CBCL face images`_ database.
        #. `ORL face images`_ database.
        #. `MovieLens 100k`_ data set.
        #. `Medlars`_ data set.
        #. `S. cerevisiae`_ MIPS Functional Catalogue data set.
        
    .. _leukemia gene expressions: http://portals.broadinstitute.org/cgi-bin/cancer/publications/pub_paper.cgi?mode=view&paper_id=89
    .. _leukemia samples: http://portals.broadinstitute.org/cgi-bin/cancer/publications/pub_paper.cgi?mode=view&paper_id=89
    .. _leukemia genes: http://portals.broadinstitute.org/cgi-bin/cancer/publications/pub_paper.cgi?mode=view&paper_id=89
    .. _medulloblastoma gene expressions: http://portals.broadinstitute.org/cgi-bin/cancer/publications/pub_paper.cgi?mode=view&paper_id=89
    .. _medulloblastoma samples: http://portals.broadinstitute.org/cgi-bin/cancer/publications/pub_paper.cgi?mode=view&paper_id=89
    .. _medulloblastoma genes: http://portals.broadinstitute.org/cgi-bin/cancer/publications/pub_paper.cgi?mode=view&paper_id=89
    .. _CBCL face images: http://cbcl.mit.edu/cbcl/software-datasets/FaceData2.html
    .. _ORL face images: http://www.cl.cam.ac.uk/research/dtg/attarchive/facedatabase.html
    .. _MovieLens 100k: http://www.grouplens.org/node/12
    .. _Medlars: http://web.eecs.utk.edu/research/lsi/
    .. _S. cerevisiae: http://dtai.cs.kuleuven.be/clus/hmc-ens/
        
"""
