
# THIS FILE IS GENERATED FROM NIMFA SETUP.PY
short_version = '1.4.0'
version = '1.4.0'
full_version = '1.4.0'
git_revision = 'eecbb05c4506ef482e788c8e6972c386dcbdb34a'
release = True

if not release:
    version = full_version
    short_version += ".dev"
